/* David Leonard, 2002. Public domain. */
/* $Id: cmine.c,v 1.10 2002/02/18 15:46:28 d Exp $ */

/*
 * Minesweeper for (n)curses.
 * User interface module.
 */

#include <curses.h>
#include <unistd.h>
#include <stdlib.h>
#include <err.h>
#include <string.h>
#include <sys/time.h>
#include "mine.h"
#include "hiscore.h"
#include "hint.h"

#define VERSION	"1.3"

static WINDOW * minewin(struct mine *, int, int);
static void	drawmine(struct mine *, WINDOW *);
static void	autoclick(struct mine *, int, int);
static void	win(struct mine *);
static void	expose(struct mine *, int, int);
static void	flag(struct mine *, int, int);
static void	mine_click(struct mine *, WINDOW *, MEVENT *);
static void	mine_key(struct mine *, int);
static void	cheat(struct mine *, WINDOW *);

/* Characters for covered cells */
static chtype untested_ch;
#define CH_BOMB		('*' | A_STANDOUT)
#define CH_FLAG		('%' | A_BOLD | COLOR_PAIR(4))
#define CH_UNKNOWN	('?' | A_BOLD)

/* Characters for uncovered cells */
static chtype tested_ch[9] = {
	(' ' | A_NORMAL),
	('1' | COLOR_PAIR(1)),			/* light blue */
	('2' | COLOR_PAIR(2)),			/* light green */
	('3' | COLOR_PAIR(3)),			/* light red */
	('4' | A_BOLD | COLOR_PAIR(1)),		/* dark blue */
	('5' | A_BOLD | COLOR_PAIR(2)),		/* dark red */
	('6' | A_BOLD | COLOR_PAIR(6)),		/* aqua */
	('7' | A_BOLD | COLOR_PAIR(7)),		/* black */
	('8' | A_BOLD | COLOR_PAIR(8)),		/* grey */
};

static int cursor_row = 0;
static int cursor_col = 0;

static struct timeval starttime;
static int started;
static int holdtime;
static int cheated;

static chtype midch[4][3] = {
	{ ':' | COLOR_PAIR(3) | A_BOLD, 
	  '-' | COLOR_PAIR(3) | A_BOLD, 
	  '(' | COLOR_PAIR(3) | A_BOLD },
	{ '[' | COLOR_PAIR(1),
	  'r' | COLOR_PAIR(1),
	  ']' | COLOR_PAIR(1) },
	{ ':' | COLOR_PAIR(2) | A_BOLD | A_REVERSE, 
	  '-' | COLOR_PAIR(2) | A_BOLD | A_REVERSE, 
	  ')' | COLOR_PAIR(2) | A_BOLD | A_REVERSE } ,
	{ ':' | COLOR_PAIR(8) | A_BOLD, 
	  '-' | COLOR_PAIR(8) | A_BOLD, 
	  ')' | COLOR_PAIR(8) | A_BOLD }
}, *mids;

/* Create a window to display the minefield & score */
static WINDOW *
minewin(m, y, x)
	struct mine *m;
	int y, x;
{
	int rows, cols;
	WINDOW *w;

	mine_size(m, &rows, &cols, NULL);
	w =  newwin(rows + 3, cols + 2, y, x);
	if (w == NULL)
		errx(1, "newwin() failed: too many rows/cols?");
	keypad(w, TRUE);
	return w;
}

/* Draw the minefield */
static void
drawmine(m, w)
	struct mine *m;
	WINDOW *w;
{
	int row, col, rows, cols;
	struct timeval now;
	int time, timelen, mw;

	/* Get the size of the minefield */
	mine_size(m, &rows, &cols, NULL);
	wborder(w, 0,0,0,0,0,0,0,0);
	if (!started)
		time = holdtime;
	else {
		gettimeofday(&now, NULL);
		timersub(&now, &starttime, &now);
		holdtime = time = now.tv_sec;
	}

	for (row = 0; row < rows; row++)
	for (col = 0; col < cols; col++) {
		int count;
		chtype c = 'x';
		switch (mine_status(m, row, col)) {
		case STATUS_UNTESTED:
			c = untested_ch;
			break;
		case STATUS_TESTED:
			count = mine_test(m, row, col);
			if (count == -1)
				c = CH_BOMB;
			else
				c = tested_ch[count];
			break;
		case STATUS_MARKED_FLAG:
			c = CH_FLAG;
			break;
		case STATUS_MARKED_UNKNOWN:
			c = CH_UNKNOWN;
			break;
		default:
			errx(1, "drawmine: mine_status() bogus");
		}
		mvwaddch(w, row + 2, col + 1, c);
	}

	timelen = time < 1000 ? 3 :
		  time < 10000 ? 4 :
		  time < 100000 ? 5 : 6;
	mvwprintw(w, 1, cols + 1 - timelen, "%0*d", timelen, time);
	mvwprintw(w, 1, 1, "%03d", mine_remaining(m));

	mw = mine_won(m);
	mids = &midch[mw == 1 && cheated ? 3 : mw + 1][0];
	mvwaddch(w, 1, cols / 2 + 0, mids[0]);
	mvwaddch(w, 1, cols / 2 + 1, mids[1]);
	mvwaddch(w, 1, cols / 2 + 2, mids[2]);

	wmove(w, cursor_row + 2, cursor_col + 1);
}

/* Automatically click on surrounding cells if we uncover a zero */
static void
autoclick(m, row, col)
	struct mine *m;
	int row, col;
{
	int rows, cols, i;

	if (mine_test(m, row, col) != 0)
		return;
	mine_size(m, &rows, &cols, NULL);
	for (i = 0; i < 8; i++)  {
		int r = row + delta[i].dx;
		int c = col + delta[i].dy;
		if (r >= 0 && r < rows && c >=0 && c < cols) {
			if (mine_status(m, r, c) != STATUS_TESTED)
				autoclick(m, r, c);
		}
	}
}

/* win: stop the timer and record score */
static void
win(m)
	struct mine *m;
{
	int rows, cols, mines, time;
	struct timeval now;

	started = 0;
	gettimeofday(&now, NULL);
	timersub(&now, &starttime, &now);
	mine_size(m, &rows, &cols, &mines);
	time = now.tv_sec;
	if (!cheated)
		hiscore(rows, cols, mines, time);
}

/* User wants to expose a (hopefully un-mined) cell */
static void
expose(m, row, col)
	struct mine *m;
	int row, col;
{
	/* Don't allow anything if we have won or lost */
	if (mine_won(m) != 0)
		return;

	/* Don't expose cells that are flagged ... Safety feature! */
	if (mine_status(m, row, col) == STATUS_MARKED_FLAG) {
		flash();
		return;
	}

	/* Test the cell... */
	if (mine_test(m, row, col) == 0)
		/* nothing under it */
		autoclick(m, row, col);

	/* Beep if we suddenly lost */
	if (!started) {
		started = 1;
		gettimeofday(&starttime, NULL);
	}
	switch (mine_won(m)) {
	case 1:
		win(m);
		break;
	case -1:
		beep();
		started = 0;
		break;
	}
}

/* User wants to change the flag status of a cell */
static void
flag(m, row, col)
	struct mine *m;
	int row, col;
{
	/* Don't allow anything if we have won or lost */
	if (mine_won(m) != 0)
		return;

	switch (mine_status(m, row, col)) {
	case STATUS_UNTESTED:
	    mine_mark(m, row, col, MARK_FLAG);
	    break;
	case STATUS_MARKED_FLAG:
	    mine_mark(m, row, col,MARK_UNKNOWN);
	    break;
	case STATUS_MARKED_UNKNOWN:
	    mine_mark(m, row, col, MARK_NONE);
	    break;
	}
	if (!started) {
		started = 1;
		gettimeofday(&starttime, NULL);
	}
	if (mine_won(m) == 1)
		win(m);
}

/* Process a user's click in the mine window */
static void
mine_click(m, w, mev)
	struct mine *m;
	WINDOW *w;
	MEVENT *mev;
{
	int rows, cols;
	int row = mev->y - 2;
	int col = mev->x - 1;

	mine_size(m, &rows, &cols, NULL);

	/* convert a click on [r] to 'r' */
	if (mev->y == 1) {
		if ((mev->x >= cols/2) && (mev->x < 3+cols/2))
			ungetch('r');
		return;
	}

	/* ignore border clicks */
	if (row < 0 || col < 0 || row >= rows || col >= cols)
		return;

	cursor_row = row;
	cursor_col = col;

	/* a look-under click */
	if (mev->bstate & BUTTON1_CLICKED) 
		expose(m, row, col);

	/* a flagging click */
	if (mev->bstate & BUTTON3_CLICKED) 
		flag(m, row, col);
}

/* Process a user's mine-action keystroke */
static void
mine_key(m, ch)
	struct mine *m;
	int ch;
{
	int rows, cols;

	mine_size(m, &rows, &cols, NULL);
	switch (ch) {
	case 'h':
	case KEY_LEFT:
		if (cursor_col > 0)
			cursor_col--;
		break;
	case 'l':
	case KEY_RIGHT:
		if (cursor_col < cols - 1)
			cursor_col++;
		break;
	case 'k':
	case KEY_UP:
		if (cursor_row > 0)
			cursor_row--;
		break;
	case 'j':
	case KEY_DOWN:
		if (cursor_row < rows - 1)
			cursor_row++;
		break;
	case '^':
	case KEY_HOME:
		cursor_col = 0;
		break;
	case '$':
	case KEY_END:
		cursor_col = cols - 1;
		break;
	case 'K':
	case KEY_PPAGE:
		if (cursor_row > 5)
			cursor_row -= 5;
		else
			cursor_row = 0;
		break;
	case 'J':
	case KEY_NPAGE:
		if (cursor_row + 5 < rows)
			cursor_row += 5;
		else
			cursor_row = rows - 1;
		break;
	case 'H':
		if (cursor_col > 5)
			cursor_col -= 5;
		else
			cursor_col = 0;
		break;
	case 'L':
		if (cursor_col + 5 < cols)
			cursor_col += 5;
		else
			cursor_col = cols - 1;
		break;
	case KEY_ENTER:
	case '\n':
	case '\r':
	case 'f':
		flag(m, cursor_row, cursor_col);
		break;
	case ' ':
		expose(m, cursor_row, cursor_col);
		break;
	default:
		flash();
		break;
	}
}

/* User wants to cheat a bit */
static void
cheat(m, w)
	struct mine *m;
	WINDOW *w;
{
	int row, col;

	for (;;) {
		switch (hint(m, &row, &col)) {
		default:
		case HINT_FAILED:
			beep();
			return;
		case HINT_EXPOSE:
			expose(m, row, col);
			cursor_row = row;
			cursor_col = col;
			break;
		case HINT_FLAG:
			flag(m, row, col);
			cursor_row = row;
			cursor_col = col;
			break;
		case HINT_TOOMANY:
			cursor_row = row;
			cursor_col = col;
			beep();
			return;
		}
		drawmine(m, w);
		wrefresh(w);
		cheated = 1;
	}
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
	struct mine *m;
	int ch;
	MEVENT mev;
	WINDOW *minescr;
	int rows, cols, mines = 0;
	int error = 0;
	int sflag = 0;
	char **arg;

	initscr();

	/* default to screen width and a mine loading of 10% */
	rows = LINES - 3;
	cols = COLS - 2;

	/* Read minefield parameters from command line */
	arg = argv + 1;
	if (*arg && (*arg)[0] == '-') {
		switch ((*arg)[1]) {
		case 'b':
			rows = 9; cols = 9; mines = 10; break;
		case 'i':
			rows = 16; cols = 16; mines = 40; break;
		case 'e':
			rows = 16; cols = 30; mines = 99; break;
		case 's':
			sflag = 1; break;
		default:
			error = 1;
		}
		if (!error && (*arg)[2] != '\0')
			error = 1;
		arg++;
	}
	if (*arg) { rows = atoi(*arg); arg++; }
	if (*arg) { cols = atoi(*arg); arg++; }
	if (*arg) { 
		mines = atoi(*arg); 
		if (strchr(*arg, '%') != NULL)
			mines = rows * cols * mines / 100;
		arg++;
	}
	if (rows <= 0 || cols <= 0 || mines < 0 || mines > rows * cols || *arg)
		error = 1;

	if (error) {
		/* Show usage */
		endwin();
		fprintf(stderr, "cmine version %s\n", VERSION);
		fprintf(stderr, "usage: %s [-sbie] [rows [cols [mines]]]\n",
		    argv[0]);
		exit(1);
	}

	if (sflag) {
		endwin();
		showscores();
		exit(0);
	}

	/* Default to 10% mines if unspecified */
	if (mines == 0)
		mines = rows * cols / 10;

	/* Initialise curses */
	use_default_colors();
	start_color();
	cbreak();
	noecho();
	nonl();
	intrflush(stdscr, FALSE);
	init_pair(1, COLOR_BLUE, -1);
	init_pair(2, COLOR_GREEN, -1);
	init_pair(3, COLOR_RED, -1);
	init_pair(4, -1, COLOR_RED);
	init_pair(6, COLOR_CYAN, -1);
	init_pair(7, COLOR_BLACK, COLOR_WHITE);
	init_pair(8, COLOR_WHITE, COLOR_BLACK);
	init_pair(9, -1, COLOR_WHITE);
	mousemask(BUTTON1_CLICKED | BUTTON3_CLICKED, NULL);

	if (has_colors())
		untested_ch = ' ' | A_BOLD | COLOR_PAIR(9);
	else
		untested_ch = A_DIM | '+';

	/* Create a new minefield */
	m = mine_new(rows, cols, mines);
	minescr = minewin(m, 0, 0);
	started = 0;
	holdtime = 0;
	cheated = 0;

	for (;;) {
		/* Draw current minefield and wait for a key */
		drawmine(m, minescr);
		wrefresh(minescr);

		if (started) {
			struct timeval now, t;
			static struct timeval one = { 1, 0 };

			/* Sleep till key or the next second increment */
			gettimeofday(&now, NULL);
			timersub(&now, &starttime, &t);
			t.tv_sec = 0;
			timersub(&one, &t, &t);
			wtimeout(minescr, t.tv_usec / 1000);
		} else
			wtimeout(minescr, -1);
		ch = wgetch(minescr);

		/* Process keypress/mouse event */
		switch (ch) {
		case ERR: /* (no key) */
			break;
		case 'q':
			goto quit;
		case 'r':
			mine_free(m);
			m = mine_new(rows, cols, mines);
			started = 0;
			holdtime = 0;
			cheated = 0;
			break;
		case 'c':
			cheat(m, minescr);
			break;
		case 'L' & 0x3f:
		case 'W' & 0x3f:
			clearok(curscr, TRUE);
			break;
		case KEY_MOUSE:
			if (getmouse(&mev) == OK) {
			    if (wenclose(minescr, mev.y, mev.x)) {
				wmouse_trafo(minescr, &mev.y, &mev.x, 0);
				mine_click(m, minescr, &mev);
			    }
			}
			break;
		default:
			mine_key(m, ch);
			break;
		}

	}
    quit:
	endwin();
	switch (mine_won(m)) {
	default:
		exit(2);
	case 1:
		if (cheated) {
			printf("I won!\n");
			exit(3);
		}
		printf("You won!\n");
		exit(0);
	case -1:
		printf("You lost!\n");
		exit(1);
	}
}
