/* David Leonard, 2002. Public domain. */
/* $Id: mine.c,v 1.4 2002/02/18 15:45:53 d Exp $ */

/*
 * Minesweeper game logic module
 */

#include <stdio.h>
#include <stdlib.h>
#include "mine.h"

struct mine {
	int	rows, cols, mines, togo, flags;
	int	revealed, exploded;
	unsigned char *map;
	unsigned char *surround;
#define MAP_MINE	1
#define MAP_TESTED	2
#define MAP_MARKED	4
#define MAP_MARK_FLAG	8	/* or question */
};

static unsigned char getcell(struct mine *, int, int);
static void setcell(struct mine *, int, int, unsigned char);
static void reveal(struct mine *);

/* Convenient table of relative coordinates */
struct delta delta[8] =
	{ {-1,-1}, {0,-1}, {+1,-1}, {-1,+0},
	  {+1,+0}, {-1,+1}, {0,+1}, {+1,+1} };

/* Return the state of a cell in the minefield */
static unsigned char
getcell(m, row, col)
	struct mine *m;
	int row, col;
{
	if (row < 0 || col < 0 || row >= m->rows || col >= m->cols)
		return 0;
	return m->map[row * m->cols + col];
}

/* Update the state of a cell in the minefield */
static void
setcell(m, row, col, value)
	struct mine *m;
	int row, col;
	unsigned char value;
{
	m->map[row * m->cols + col] = value;
}

/* Set the state of all mines to be exposed */
static void
reveal(m)
	struct mine *m;
{
	int i;
	int max = m->rows * m->cols;

	if (m->revealed)
		return;
	for (i = 0; i < max; i++)
		if (m->map[i] & MAP_MINE)
			m->map[i] |= MAP_TESTED;
	m->revealed = 1;
}

/* Create a new minefield */
struct mine *
mine_new(rows, cols, mines)
	int rows, cols, mines;
{
	struct mine *m;
	int i, p, row, col;

	m = (struct mine *)malloc(sizeof (struct mine));
	if (m == NULL)
		return NULL;
	m->rows = rows;
	m->cols = cols;
	m->mines = mines;
	m->revealed = 0;
	m->map = (unsigned char *)malloc(rows * cols * sizeof (unsigned char));
	if (m->map == NULL) {
		free(m);
		return NULL;
	}
	m->surround = (unsigned char *)malloc(rows * cols
	    * sizeof (unsigned char));
	if (m->surround == NULL) {
		free(m->map);
		free(m);
		return NULL;
	}
	/* Place the mines */
	srandomdev();
	for (i = 0; i < rows * cols; i ++)
		m->map[i] = 0;
	p = 0;
	for (i = 0; i < mines; i++) {
		int q;
		for (q = random() % (rows * cols); q; q--) {
			p = (p+1) % (rows * cols);
			while (m->map[p] == MAP_MINE)
				p = (p+1) % (rows * cols);
		}
		while (m->map[p] == MAP_MINE)
			p = (p+1) % (rows * cols);
		/*
		 * XXX - by allowing mines to be placed on the
		 * edge of the minefield, we run the risk of
		 * having games that require a guess! (First move or two
		 * are always a guess anyway)
		 */
		m->map[p] = MAP_MINE;
	}

	/* Pre-compute all the surround counts */
	for (row = 0; row < rows; row++)
	    for (col = 0; col < cols; col++) {
		int count = 0;
		for (i = 0; i < 8; i++)
			if (getcell(m, row + delta[i].dx, col + delta[i].dy)
			    & MAP_MINE)
				count++;
		m->surround[row * m->cols + col] = count;
	    }

	m->togo = rows * cols - mines;
	m->exploded = 0;
	m->flags = 0;

	return m;
}

/* Deallocate storage with a minefield */
void
mine_free(m)
	struct mine *m;
{
	free(m->map);
	free(m);
}

/* See what is under a cell (may explode!) */
int
mine_test(m, row, col)
	struct mine *m;
	int row, col;
{
	unsigned char status = getcell(m, row, col);

	if ((status & MAP_TESTED) == 0) {
		status |= MAP_TESTED;
		if (status & MAP_MARK_FLAG)
			m->flags--;
		status &= ~(MAP_MARKED | MAP_MARK_FLAG);
		setcell(m, row, col, status);
		if (status & MAP_MINE) {
			m->exploded = 1;
			reveal(m);
		} else
			m->togo--;
	}
	if (status & MAP_MINE)
		return -1;
	return m->surround[row * m->cols + col];
}

/* Return the state of a cell without testing it */
int
mine_status(m, row, col)
	struct mine *m;
	int row, col;
{
	unsigned char status = getcell(m, row, col);
	if ((status & (MAP_MARKED|MAP_TESTED)) == MAP_MARKED) 
		return (status & MAP_MARK_FLAG) ? STATUS_MARKED_FLAG 
						: STATUS_MARKED_UNKNOWN;
	return (status & MAP_TESTED) ? STATUS_TESTED
				     : STATUS_UNTESTED;
}

/* Set a mark on a minefield cell */
void
mine_mark(m, row, col, type)
	struct mine *m;
	int row, col, type;
{
	unsigned char status = getcell(m, row, col);
	if (status & MAP_TESTED)
		return;
	if (status & MAP_MARK_FLAG)
		m->flags--;
	status &= ~ (MAP_MARKED|MAP_MARK_FLAG);
	switch (type) {
	case MARK_NONE:
		break;
	case MARK_FLAG:
		status |= MAP_MARKED|MAP_MARK_FLAG;
		m->flags++;
		break;
	case MARK_UNKNOWN:
		status |= MAP_MARKED;
		break;
	}
	setcell(m, row, col, status);
}

/* number of actual mines minus minus number of flags set */
int
mine_remaining(m)
	struct mine *m;
{
	return m->mines - m->flags;
}

/* Figure out if the player has won or lost (or neither) */
int
mine_won(m)
	struct mine *m;
{
	if (m->exploded)
		return -1;
	if (m->togo == 0)
		return 1;
	return 0;
}

/* Return the game parameters */
void
mine_size(m, rows, cols, mines)
	struct mine *m;
	int *rows, *cols, *mines;
{
	if (rows)
		*rows = m->rows;
	if (cols)
		*cols = m->cols;
	if (mines)
		*mines = m->mines;
}
